﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2015 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using HeuristicLab.Analysis;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Optimization.Operators;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Random;

namespace HeuristicLab.Algorithms.ALPS {
  [Item("ALPS", "Base class for all ALPS-based algorithms.")]
  [StorableClass]
  public abstract class Alps : HeuristicOptimizationEngineAlgorithm, IStorableContent {
    public string Filename { get; set; }

    #region Problem Properties
    public override Type ProblemType {
      get { return typeof(ISingleObjectiveHeuristicOptimizationProblem); }
    }
    public new ISingleObjectiveHeuristicOptimizationProblem Problem {
      get { return (ISingleObjectiveHeuristicOptimizationProblem)base.Problem; }
      set { base.Problem = value; }
    }
    #endregion

    #region Parameter Properties
    private IValueParameter<IntValue> SeedParameter {
      get { return (IValueParameter<IntValue>)Parameters["Seed"]; }
    }
    private IValueParameter<BoolValue> SetSeedRandomlyParameter {
      get { return (IValueParameter<BoolValue>)Parameters["SetSeedRandomly"]; }
    }
    private IFixedValueParameter<MultiAnalyzer> AnalyzerParameter {
      get { return (IFixedValueParameter<MultiAnalyzer>)Parameters["Analyzer"]; }
    }
    private IFixedValueParameter<MultiAnalyzer> LayerAnalyzerParameter {
      get { return (IFixedValueParameter<MultiAnalyzer>)Parameters["LayerAnalyzer"]; }
    }
    private IValueParameter<IntValue> NumberOfLayersParameter {
      get { return (IValueParameter<IntValue>)Parameters["NumberOfLayers"]; }
    }
    private IValueParameter<EnumValue<AgingScheme>> AgingSchemeParameter {
      get { return (IValueParameter<EnumValue<AgingScheme>>)Parameters["AgingScheme"]; }
    }
    private IValueParameter<IntValue> AgeGapParameter {
      get { return (IValueParameter<IntValue>)Parameters["AgeGap"]; }
    }
    private IValueParameter<IntArray> AgeLimitsParameter {
      get { return (IValueParameter<IntArray>)Parameters["AgeLimits"]; }
    }
    private IValueParameter<EnumValue<AgeInheritance>> AgeInheritanceParameter {
      get { return (IValueParameter<EnumValue<AgeInheritance>>)Parameters["AgeInheritance"]; }
    }
    private IValueParameter<ReductionOperation> AgeInheritanceReductionParameter {
      get { return (IValueParameter<ReductionOperation>)Parameters["AgeInheritanceReduction"]; }
    }
    private IValueParameter<IntValue> MatingPoolRangeParameter {
      get { return (IValueParameter<IntValue>)Parameters["MatingPoolRange"]; }
    }
    private IValueParameter<PercentValue> MatingPoolSelectionPercentageParameter {
      get { return (IValueLookupParameter<PercentValue>)Parameters["MatingPoolSelectionPercentage"]; }
    }
    #endregion

    #region Properties
    public IntValue Seed {
      get { return SeedParameter.Value; }
      set { SeedParameter.Value = value; }
    }
    public BoolValue SetSeedRandomly {
      get { return SetSeedRandomlyParameter.Value; }
      set { SetSeedRandomlyParameter.Value = value; }
    }
    public MultiAnalyzer Analyzer {
      get { return AnalyzerParameter.Value; }
    }
    public MultiAnalyzer LayerAnalyzer {
      get { return LayerAnalyzerParameter.Value; }
    }
    public IntValue NumberOfLayers {
      get { return NumberOfLayersParameter.Value; }
      set { NumberOfLayersParameter.Value = value; }
    }
    public EnumValue<AgingScheme> AgingScheme {
      get { return AgingSchemeParameter.Value; }
      set { AgingSchemeParameter.Value = value; }
    }
    public IntValue AgeGap {
      get { return AgeGapParameter.Value; }
      set { AgeGapParameter.Value = value; }
    }
    public IntArray AgeLimits {
      get { return AgeLimitsParameter.Value; }
      set { AgeLimitsParameter.Value = value; }
    }
    public EnumValue<AgeInheritance> AgeInheritance {
      get { return AgeInheritanceParameter.Value; }
      set { AgeInheritanceParameter.Value = value; }
    }
    private ReductionOperation AgeInheritanceReduction {
      get { return AgeInheritanceReductionParameter.Value; }
      set { AgeInheritanceReductionParameter.Value = value; }
    }
    public IntValue MatingPoolRange {
      get { return MatingPoolRangeParameter.Value; }
      set { MatingPoolRangeParameter.Value = value; }
    }
    public PercentValue MatingPoolSelectionPercentage {
      get { return MatingPoolSelectionPercentageParameter.Value; }
      set { MatingPoolSelectionPercentageParameter.Value = value; }
    }
    #endregion

    #region Helper Properties
    protected RandomCreator GlobalRandomCreator {
      get { return (RandomCreator)OperatorGraph.InitialOperator; }
    }
    protected SolutionsCreator SolutionsCreator {
      get { return OperatorGraph.Iterate().OfType<SolutionsCreator>().First(); }
    }
    #endregion

    #region Preconfigured Analyzers
    [Storable]
    private BestAverageWorstQualityAnalyzer qualityAnalyzer;
    [Storable]
    private BestAverageWorstQualityAnalyzer layerQualityAnalyzer;
    #endregion

    [StorableConstructor]
    protected Alps(bool deserializing)
      : base(deserializing) { }
    protected Alps(Alps original, Cloner cloner)
      : base(original, cloner) {
      qualityAnalyzer = cloner.Clone(original.qualityAnalyzer);
      layerQualityAnalyzer = cloner.Clone(original.layerQualityAnalyzer);
      Initialize();
    }

    protected Alps()
      : base() {
      Parameters.Add(new ValueParameter<IntValue>("Seed", "The random seed used to initialize the new pseudo random number generator.", new IntValue(0)));
      Parameters.Add(new ValueParameter<BoolValue>("SetSeedRandomly", "True if the random seed should be set to a random value, otherwise false.", new BoolValue(true)));
      Parameters.Add(new FixedValueParameter<MultiAnalyzer>("Analyzer", "The operator used to analyze the islands.", new MultiAnalyzer()));
      Parameters.Add(new FixedValueParameter<MultiAnalyzer>("LayerAnalyzer", "The operator used to analyze each layer.", new MultiAnalyzer()));
      Parameters.Add(new ValueParameter<IntValue>("NumberOfLayers", "The number of layers.", new IntValue(10)));
      Parameters.Add(new ValueParameter<EnumValue<AgingScheme>>("AgingScheme", "The aging scheme for setting the age-limits for the layers.", new EnumValue<AgingScheme>(ALPS.AgingScheme.Polynomial)));
      Parameters.Add(new ValueParameter<IntValue>("AgeGap", "The frequency of reseeding the lowest layer and scaling factor for the age-limits for the layers", new IntValue(20)));
      Parameters.Add(new ValueParameter<IntArray>("AgeLimits", new IntArray(new int[0])) { Hidden = true });
      Parameters.Add(new ValueParameter<EnumValue<AgeInheritance>>("AgeInheritance", "The operator for determining the age of an offspring based the parents' age.", new EnumValue<AgeInheritance>(ALPS.AgeInheritance.Older)));
      Parameters.Add(new ValueParameter<ReductionOperation>("AgeInheritanceReduction") { Hidden = true });
      Parameters.Add(new ValueParameter<IntValue>("MatingPoolRange", "The range of layers used for creating a mating pool. (1 = current + previous layer)", new IntValue(1)) { Hidden = true });
      Parameters.Add(new ValueParameter<PercentValue>("MatingPoolSelectionPercentage", "Percentage of the previous layers used for creating a mating pool.", new PercentValue(1.0, restrictToUnitInterval: true)) { Hidden = true });

      qualityAnalyzer = new BestAverageWorstQualityAnalyzer();
      layerQualityAnalyzer = new BestAverageWorstQualityAnalyzer();

      ParameterizeAgeLimits();
      ParameterizeAgeInheritanceReduction();

      ParameterizeAnalyzers();
      UpdateAnalyzers();

      Initialize();
    }


    #region Events
    public override void Prepare() {
      if (Problem != null)
        base.Prepare();
    }

    protected override void OnProblemChanged() {
      base.OnProblemChanged();
      ParameterizeStochasticOperator(Problem.SolutionCreator);
      ParameterizeStochasticOperatorForLayer(Problem.Evaluator);
      foreach (var @operator in Problem.Operators.OfType<IOperator>())
        ParameterizeStochasticOperator(@operator);
      ParameterizeSolutionsCreator();
      ParameterizeAnalyzers();
      UpdateAnalyzers();
      Problem.Evaluator.QualityParameter.ActualNameChanged += Evaluator_QualityParameter_ActualNameChanged;
    }
    protected override void Problem_SolutionCreatorChanged(object sender, EventArgs e) {
      base.Problem_SolutionCreatorChanged(sender, e);
      ParameterizeStochasticOperator(Problem.SolutionCreator);
      ParameterizeSolutionsCreator();
    }
    protected override void Problem_EvaluatorChanged(object sender, EventArgs e) {
      base.Problem_EvaluatorChanged(sender, e);
      ParameterizeStochasticOperatorForLayer(Problem.Evaluator);
      ParameterizeSolutionsCreator();
      ParameterizeAnalyzers();
      Problem.Evaluator.QualityParameter.ActualNameChanged += Evaluator_QualityParameter_ActualNameChanged;
    }

    protected override void Problem_OperatorsChanged(object sender, EventArgs e) {
      foreach (var @operator in Problem.Operators.OfType<IOperator>())
        ParameterizeStochasticOperator(@operator);
      UpdateAnalyzers();
      base.Problem_OperatorsChanged(sender, e);
    }
    protected virtual void Evaluator_QualityParameter_ActualNameChanged(object sender, EventArgs e) {
      ParameterizeAnalyzers();
    }

    private void AgeGapParameter_ValueChanged(object sender, EventArgs e) {
      AgeGap.ValueChanged += AgeGap_ValueChanged;
      ParameterizeAgeLimits();
    }
    private void AgeGap_ValueChanged(object sender, EventArgs e) {
      ParameterizeAgeLimits();
    }
    private void AgingSchemeParameter_ValueChanged(object sender, EventArgs e) {
      AgingScheme.ValueChanged += AgingScheme_ValueChanged;
      ParameterizeAgeLimits();
    }
    private void AgingScheme_ValueChanged(object sender, EventArgs e) {
      ParameterizeAgeLimits();
    }
    private void NumberOfLayersParameter_ValueChanged(object sender, EventArgs e) {
      NumberOfLayers.ValueChanged += NumberOfLayers_ValueChanged;
      ParameterizeAgeLimits();
    }
    private void NumberOfLayers_ValueChanged(object sender, EventArgs e) {
      ParameterizeAgeLimits();
    }
    private void AgeInheritanceParameter_ValueChanged(object sender, EventArgs e) {
      AgeInheritance.ValueChanged += AgeInheritance_ValueChanged;
      ParameterizeAgeInheritanceReduction();
    }
    private void AgeInheritance_ValueChanged(object sender, EventArgs e) {
      ParameterizeAgeInheritanceReduction();
    }
    private void AnalyzerOperators_ItemsAdded(object sender, CollectionItemsChangedEventArgs<IndexedItem<IAnalyzer>> e) {
      foreach (var analyzer in e.Items) {
        foreach (var parameter in analyzer.Value.Parameters.OfType<IScopeTreeLookupParameter>()) {
          parameter.Depth = 2;
        }
      }
    }
    private void LayerAnalyzerOperators_ItemsAdded(object sender, CollectionItemsChangedEventArgs<IndexedItem<IAnalyzer>> e) {
      foreach (var analyzer in e.Items) {
        IParameter resultParameter;
        if (analyzer.Value.Parameters.TryGetValue("Results", out resultParameter)) {
          var lookupParameter = resultParameter as ILookupParameter;
          if (lookupParameter != null)
            lookupParameter.ActualName = "LayerResults";
        }
        foreach (var parameter in analyzer.Value.Parameters.OfType<IScopeTreeLookupParameter>()) {
          parameter.Depth = 1;
        }
      }
    }
    #endregion

    #region Parameterization
    private void Initialize() {
      if (Problem != null)
        Problem.Evaluator.QualityParameter.ActualNameChanged += Evaluator_QualityParameter_ActualNameChanged;
      AgeGapParameter.ValueChanged += AgeGapParameter_ValueChanged;
      AgeGap.ValueChanged += AgeGap_ValueChanged;
      AgingSchemeParameter.ValueChanged += AgingSchemeParameter_ValueChanged;
      AgingScheme.ValueChanged += AgingScheme_ValueChanged;
      NumberOfLayersParameter.ValueChanged += NumberOfLayersParameter_ValueChanged;
      NumberOfLayers.ValueChanged += NumberOfLayers_ValueChanged;
      AgeInheritanceParameter.ValueChanged += AgeInheritanceParameter_ValueChanged;
      AgeInheritance.ValueChanged += AgeInheritance_ValueChanged;
      Analyzer.Operators.ItemsAdded += AnalyzerOperators_ItemsAdded;
      LayerAnalyzer.Operators.ItemsAdded += LayerAnalyzerOperators_ItemsAdded;
    }
    private void ParameterizeSolutionsCreator() {
      SolutionsCreator.EvaluatorParameter.ActualName = Problem.EvaluatorParameter.Name;
      SolutionsCreator.SolutionCreatorParameter.ActualName = Problem.SolutionCreatorParameter.Name;
    }
    private void ParameterizeAnalyzers() {
      qualityAnalyzer.ResultsParameter.ActualName = "Results";
      qualityAnalyzer.ResultsParameter.Hidden = true;
      qualityAnalyzer.QualityParameter.Depth = 2;
      layerQualityAnalyzer.ResultsParameter.ActualName = "Results";
      layerQualityAnalyzer.ResultsParameter.Hidden = true;
      layerQualityAnalyzer.QualityParameter.Depth = 1;
      if (Problem != null) {
        qualityAnalyzer.MaximizationParameter.ActualName = Problem.MaximizationParameter.Name;
        qualityAnalyzer.MaximizationParameter.Hidden = true;
        qualityAnalyzer.QualityParameter.ActualName = Problem.Evaluator.QualityParameter.ActualName;
        qualityAnalyzer.QualityParameter.Hidden = true;
        qualityAnalyzer.BestKnownQualityParameter.ActualName = Problem.BestKnownQualityParameter.Name;
        qualityAnalyzer.BestKnownQualityParameter.Hidden = true;
        layerQualityAnalyzer.MaximizationParameter.ActualName = Problem.MaximizationParameter.Name;
        layerQualityAnalyzer.MaximizationParameter.Hidden = true;
        layerQualityAnalyzer.QualityParameter.ActualName = Problem.Evaluator.QualityParameter.ActualName;
        layerQualityAnalyzer.QualityParameter.Hidden = true;
        layerQualityAnalyzer.BestKnownQualityParameter.ActualName = Problem.BestKnownQualityParameter.Name;
        layerQualityAnalyzer.BestKnownQualityParameter.Hidden = true;
      }
    }

    private void ParameterizeAgeInheritanceReduction() {
      AgeInheritanceReduction.Value = GetAgeInheritanceReduction(AgeInheritance.Value);
    }
    protected abstract ReductionOperations GetAgeInheritanceReduction(AgeInheritance ageInheritance);
    private void ParameterizeAgeLimits() {
      var scheme = AgingScheme.Value;
      int ageGap = AgeGap.Value;
      int numberOfLayers = NumberOfLayers.Value;
      AgeLimits = scheme.CalculateAgeLimits(ageGap, numberOfLayers);
    }
    protected virtual void ParameterizeStochasticOperator(IOperator @operator) {
      var stochasticOperator = @operator as IStochasticOperator;
      if (stochasticOperator != null) {
        stochasticOperator.RandomParameter.ActualName = "GlobalRandom";
        stochasticOperator.RandomParameter.Hidden = true;
      }
    }
    protected virtual void ParameterizeStochasticOperatorForLayer(IOperator @operator) {
      var stochasticOperator = @operator as IStochasticOperator;
      if (stochasticOperator != null) {
        stochasticOperator.RandomParameter.ActualName = "LocalRandom";
        stochasticOperator.RandomParameter.Hidden = true;
      }
    }
    #endregion

    #region Updates
    private void UpdateAnalyzers() {
      Analyzer.Operators.Clear();
      LayerAnalyzer.Operators.Clear();

      Analyzer.Operators.Add(qualityAnalyzer, qualityAnalyzer.EnabledByDefault);
      LayerAnalyzer.Operators.Add(layerQualityAnalyzer, layerQualityAnalyzer.EnabledByDefault);

      if (Problem != null) {
        foreach (var analyzer in Problem.Operators.OfType<IAnalyzer>()) {
          Analyzer.Operators.Add(analyzer, analyzer.EnabledByDefault);
        }
      }
    }
    #endregion
  }
}